"""
Helper functions for config.
"""
import os
from pathlib import Path
from typing import Optional

from ruamel.yaml import YAML
from ruamel.yaml.constructor import SafeConstructor


_CONFIGURATION_PATH: Optional[Path] = None


def default_hass_config_dir():
    """Put together the default configuration directory based on the OS."""
    data_dir = os.getenv("APPDATA") if os.name == "nt" else os.path.expanduser("~")
    return os.path.join(data_dir, ".homeassistant")


def find_hass_config():
    """Try to find HASS config."""
    if "HASSIO_TOKEN" in os.environ:
        return "/config"

    config_dir = default_hass_config_dir()

    if os.path.isdir(config_dir):
        return config_dir

    raise ValueError(
        "Unable to automatically find the location of Home Assistant "
        "config. Please pass it in."
    )


class HassSafeConstructor(SafeConstructor):
    """Hass specific SafeConstructor."""


def _secret_yaml(loader, node):
    """Load secrets and embed it into the configuration YAML."""
    dirpath = Path(os.path.dirname(loader.name))

    # Recursively search for the secrets.yaml file
    # this might be needed when a yaml file is included like
    # `!included folder/file.yaml`. Same rules as
    # https://www.home-assistant.io/docs/configuration/secrets/#debugging-secrets
    fname = dirpath / "secrets.yaml"
    for _ in range(len(dirpath.parts)):
        if fname.exists() or (
            _CONFIGURATION_PATH is not None
            and _CONFIGURATION_PATH.exists()
            and dirpath.samefile(_CONFIGURATION_PATH)
        ):
            break
        dirpath = dirpath.parent
        fname = dirpath / "secrets.yaml"

    try:
        with fname.open() as secret_file:
            secrets = YAML(typ="safe").load(secret_file)
    except FileNotFoundError:
        raise ValueError("Secrets file {} not found".format(fname)) from None

    try:
        return secrets[node.value]
    except KeyError:
        raise ValueError("Secret {} not found".format(node.value)) from None


def _include_yaml(loader, node):
    """Load another YAML file and embeds it using the !include tag.

    Example:
        device_tracker: !include device_tracker.yaml
    """
    dirpath = Path(os.path.dirname(loader.name))
    fname = dirpath / node.value
    for _ in range(len(dirpath.parts)):
        if fname.exists() or (
            _CONFIGURATION_PATH is not None
            and _CONFIGURATION_PATH.exists()
            and dirpath.samefile(_CONFIGURATION_PATH)
        ):
            break
        dirpath = dirpath.parent
        fname = dirpath / node.value

    return load_yaml(os.path.join(dirpath, node.value))


def _stub_tag(constructor, node):
    """Stub a constructor with a dictionary."""
    seen = getattr(constructor, "_stub_seen", None)

    if seen is None:
        seen = constructor._stub_seen = set()

    if node.tag not in seen:
        print("YAML tag {} is not supported".format(node.tag))
        seen.add(node.tag)

    return {}


HassSafeConstructor.add_constructor("!include", _include_yaml)
HassSafeConstructor.add_constructor("!env_var", _stub_tag)
HassSafeConstructor.add_constructor("!secret", _secret_yaml)
HassSafeConstructor.add_constructor("!include_dir_list", _stub_tag)
HassSafeConstructor.add_constructor("!include_dir_merge_list", _stub_tag)
HassSafeConstructor.add_constructor("!include_dir_named", _stub_tag)
HassSafeConstructor.add_constructor("!include_dir_merge_named", _stub_tag)


def load_hass_config(path):
    """Load the HASS config."""
    return load_yaml(os.path.join(path, "configuration.yaml"))


def load_yaml(fname):
    """Load a YAML file."""
    yaml = YAML(typ="safe")
    # Compat with HASS
    yaml.allow_duplicate_keys = True
    # Stub HASS constructors
    HassSafeConstructor.name = fname
    yaml.Constructor = HassSafeConstructor

    with open(fname, encoding="utf-8") as conf_file:
        # If configuration file is empty YAML returns None
        # We convert that to an empty dict
        return yaml.load(conf_file) or {}


def db_url_from_hass_config(path):
    """Find the recorder database url from a HASS config dir."""
    global _CONFIGURATION_PATH
    _CONFIGURATION_PATH = Path(path).resolve()
    config = load_hass_config(path)
    default_path = os.path.join(path, "home-assistant_v2.db")
    default_url = "sqlite:///{}".format(default_path)

    recorder = config.get("recorder")

    if recorder:
        db_url = recorder.get("db_url")
        if db_url is not None:
            return db_url

    if not os.path.isfile(default_path):
        raise ValueError(
            "Unable to determine DB url from hass config at {}".format(path)
        )

    return default_url



"""Auth helper."""
import json
import os

from . import config


def auth_from_hass_config(path=None, **kwargs):
    """Initialize auth from HASS config."""
    if path is None:
        path = config.find_hass_config()

    return Auth(os.path.join(path, ".storage/auth"), **kwargs)


class Auth:
    """Class to hold auth."""

    def __init__(self, auth_path):
        """Load auth data and store in class."""
        with open(auth_path) as fp:
            auth = json.load(fp)

        self.users = {user["id"]: user for user in auth["data"]["users"]}
        self.refresh_tokens = {
            token["id"]: {
                "id": token["id"],
                "user": self.users[token["user_id"]],
                "client_name": token["client_name"],
                "client_id": token["client_id"],
            }
            for token in auth["data"]["refresh_tokens"]
        }

    def user_name(self, user_id):
        """Return name for user."""
        user = self.users.get(user_id)

        if user is None:
            return "Unknown user ({})".format(user_id)

        return user["name"]



"""
Helper functions.
"""
import json
import pandas as pd


def format_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """Convert states to numeric where possible and format the last_changed."""
    df["state"] = pd.to_numeric(df["state"], errors="coerce")
    df["last_updated"] = pd.to_datetime(
        df["last_updated"].values, errors="ignore", utc=True
    ).tz_localize(None)
    df = df.dropna()
    return df



"""
Classes and functions for parsing home-assistant data.
"""
from typing import Tuple

from urllib.parse import urlparse
import pandas as pd
from sqlalchemy import create_engine, text

from . import config, functions


def db_from_hass_config(path=None, **kwargs):
    """Initialize a database from HASS config."""
    if path is None:
        path = config.find_hass_config()

    url = config.db_url_from_hass_config(path)
    return HassDatabase(url, **kwargs)


def get_db_type(url):
    return urlparse(url).scheme.split("+")[0]


def stripped_db_url(url):
    """Return a version of the DB url with the password stripped out."""
    parsed = urlparse(url)

    if parsed.password is None:
        return url

    return parsed._replace(
        netloc="{}:***@{}".format(parsed.username, parsed.hostname)
    ).geturl()


class HassDatabase:
    """
    Initializing the parser fetches all of the data from the database and
    places it in a master pandas dataframe.
    """

    def __init__(self, url, *, fetch_entities=True):
        """
        Parameters
        ----------
        url : str
            The URL to the database.
        """
        self.url = url
        self.entities = None
        try:
            self.engine = create_engine(url)
            print("Successfully connected to database", stripped_db_url(url))
            if fetch_entities:
                self.fetch_entities()
        except Exception as exc:
            if isinstance(exc, ImportError):
                raise RuntimeError(
                    "The right dependency to connect to your database is "
                    "missing. Please make sure that it is installed."
                )

            print(exc)
            raise

        self.db_type = get_db_type(url)

    def perform_query(self, query, **params):
        """Perform a query."""
        try:
            return self.engine.execute(query, params)
        except:
            print(f"Error with query: {query}")
            raise

    def fetch_entities(self) -> None:
        """Fetch entities for which we have data."""
        query = text(
            """
            SELECT DISTINCT(entity_id) FROM states
            """
        )
        response = self.perform_query(query)

        # Parse the domains from the entities.
        self.entities = [e[0] for e in response]
        print(f"There are {len(self.entities)} entities with data")

    def fetch_all_sensor_data(self, limit=50000, get_attributes=False) -> pd.DataFrame:
        """
        Fetch data for all sensor entities.

        Arguments:
        - limit (default: 50000): Limit the maximum number of state changes loaded.
            If None, there is no limit.
        - get_attributes: If True, LEFT JOIN the attributes table to retrieve event's attributes.
        """
        
        if get_attributes:
            query = """
                SELECT entity_id, state, last_updated, shared_attrs
            """
        else:
            query = """
                SELECT entity_id, state, last_updated
            """
        
        query += "FROM states"
        
        if get_attributes:
            query += """
                LEFT JOIN state_attributes ON states.attributes_id = state_attributes.attributes_id
            """
            
        query += """
            WHERE
                entity_id  LIKE '%sensor%'
            AND
                state NOT IN ('unknown', 'unavailable')
            ORDER BY last_updated DESC
            """
        
        if limit is not None:
            query += f"LIMIT {limit}"
        df = pd.read_sql_query(query, self.url)
        print(f"The returned Pandas dataframe has {df.shape[0]} rows of data.")
        return df

    def fetch_all_data_of(self, sensors: Tuple[str], limit=50000, get_attributes=False) -> pd.DataFrame:
        """
        Fetch data for sensors.

        Arguments:
        - limit (default: 50000): Limit the maximum number of state changes loaded.
            If None, there is no limit.
        - get_attributes: If True, LEFT JOIN the attributes table to retrieve event's attributes.
        """
        sensors_str = str(tuple(sensors))
        if len(sensors) == 1:
            sensors_str = sensors_str.replace(",", "")

        if get_attributes:
            query = """
                SELECT entity_id, state, last_updated, shared_attrs
            """
        else:
            query = """
                SELECT entity_id, state, last_updated
            """

        query += "FROM states"
        
        if get_attributes:
            query += """
                LEFT JOIN state_attributes ON states.attributes_id = state_attributes.attributes_id
            """
            
        query += f"""
            WHERE
                entity_id IN {sensors_str}
            AND
                state NOT IN ('unknown', 'unavailable')
            ORDER BY last_updated DESC
            """

        if limit is not None:
            query += f"LIMIT {limit}"
            
        df = pd.read_sql_query(query, self.url)
        print(f"The returned Pandas dataframe has {df.shape[0]} rows of data.")
        return df



from setuptools import setup, find_packages

REQUIRES = [
    "pandas>=1.1.4",
    "ruamel.yaml>=0.15.78",
    "SQLAlchemy>=1.2.8",
    "pytz",
]

PROJECT_DESCRIPTION = "Tools for studying Home Assistant data."
PROJECT_LONG_DESCRIPTION = (
    "Home Assistant is an open-source "
    "home automation platform running on Python 3. "
    "This package provides a set of convenience "
    "functions and classes to analyse the data "
    "in your Home Assistant database. "
)

setup(
    name="HASS-data-detective",
    version="2.6",
    packages=find_packages(exclude=("tests",)),
    url="https://github.com/robmarkcole/HASS-data-detective",
    keywords=["home", "automation"],
    author="Robin Cole",
    author_email="robmarkcole@gmail.com",
    description=PROJECT_DESCRIPTION,
    long_description=PROJECT_LONG_DESCRIPTION,
    install_requires=REQUIRES,
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)



"""
Helper functions for datetimes.
"""
from datetime import datetime
import time

import pytz

UTC = pytz.UTC
# Ordered list of time categories that `time_category` produces
TIME_CATEGORIES = ["morning", "daytime", "evening", "night"]

# To localize the returned UTC times to local times
LOCAL_UTC_OFFSET = datetime.fromtimestamp(time.time()) - datetime.utcfromtimestamp(
    time.time()
)


def localize(dt):
    """Localize a datetime object to local time."""
    if dt.tzinfo is UTC:
        return (dt + LOCAL_UTC_OFFSET).replace(tzinfo=None)
    # No TZ info so not going to assume anything, return as-is.
    return dt


def is_weekday(dtObj):
    """Check a datetime object dtObj is a weekday"""
    return dtObj.weekday() < 5


def time_category(dtObj):
    """Return a time category, bed, home, work, given a dtObj."""
    if 9 <= dtObj.hour <= 17:
        return "daytime"
    elif 5 <= dtObj.hour < 9:
        return "morning"
    elif 17 < dtObj.hour < 23:
        return "evening"
    else:
        return "night"


def sqlalch_datetime(dt):
    """Convert a SQLAlchemy datetime string to a datetime object."""
    if isinstance(dt, str):
        return datetime.strptime(dt, "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=UTC)
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None:
        return dt.astimezone(UTC)
    return dt.replace(tzinfo=UTC)
